@@ -7,15 +7,16 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
+import re
 from abc import ABCMeta
+from copy import copy
 from itertools import count
 from typing import TYPE_CHECKING, cast, overload, Any, Dict, List, Iterator, \
     Optional, Union, Tuple, Type, MutableMapping, MutableSequence
 from elementpath import XPathContext, XPath2Parser, build_node_tree, protocols
 from elementpath.etree import etree_tostring
 
-from .exceptions import XMLSchemaAttributeError, XMLSchemaKeyError, XMLSchemaTypeError, \
-    XMLSchemaValueError
+from .exceptions import XMLSchemaAttributeError, XMLSchemaTypeError, XMLSchemaValueError
 from .aliases import ElementType, XMLSourceType, NamespacesType, BaseXsdType, DecodeType
 from .helpers import get_namespace, get_prefixed_qname, local_name, raw_xml_encode
 from .converters import ElementData, XMLSchemaConverter
@@ -137,28 +138,30 @@ class DataElement(MutableSequence['DataElement']):
         """The string value of the data element."""
         return raw_xml_encode(self.value)
 
-    def _map_attribute_prefixed_name(self, name: str) -> str:
-        try:
-            prefix, _local_name = name.split(':')
-        except ValueError:
-            raise XMLSchemaValueError(f'{name!r} has a wrong format') from None
-        else:
-            try:
-                return '{%s}%s' % (self.nsmap[prefix], _local_name)
-            except KeyError:
-                msg = f'prefix {prefix!r} not found in {self} nsmap'
-                raise XMLSchemaKeyError(msg) from None
-
     def get(self, key: str, default: Any = None) -> Any:
         """Gets a data element attribute."""
-        if not key.startswith('{') and ':' in key:
-            key = self._map_attribute_prefixed_name(key)
-        return self.attrib.get(key, default)
+        try:
+            return self.attrib[key]
+        except KeyError:
+            if not self.nsmap:
+                return default
+
+            # Try a match with mapped/unmapped name
+            if key.startswith('{'):
+                key = get_prefixed_qname(key, self.nsmap)
+                return self.attrib.get(key, default)
+            elif ':' in key:
+                try:
+                    _prefix, _local_name = key.split(':')
+                    key = '{%s}%s' % (self.nsmap[_prefix], _local_name)
+                except (ValueError, KeyError):
+                    pass
+                else:
+                    return self.attrib.get(key, default)
+            return default
 
     def set(self, key: str, value: Any) -> None:
         """Sets a data element attribute."""
-        if not key.startswith('{') and ':' in key:
-            key = self._map_attribute_prefixed_name(key)
         self.attrib[key] = value
 
     @property
@@ -187,6 +190,71 @@ class DataElement(MutableSequence['DataElement']):
         """The local part of the tag."""
         return local_name(self.tag)
 
+    def iter(self, tag: Optional[str] = None) -> Iterator['DataElement']:
+        """
+        Creates an iterator for the data element and its subelements. If tag
+        is not `None` or '*', only data elements whose matches tag are returned
+        from the iterator.
+        """
+        if tag == '*':
+            tag = None
+        if tag is None or tag == self.tag:
+            yield self
+        for child in self._children:
+            yield from child.iter(tag)
+
+    def iterchildren(self, tag: Optional[str] = None) -> Iterator['DataElement']:
+        """
+        Creates an iterator for the child data elements. If *tag* is not `None` or '*',
+        only data elements whose name matches tag are returned from the iterator.
+        """
+        if tag == '*':
+            tag = None
+        for child in self:
+            if tag is None or tag == child.tag:
+                yield child
+
+    def get_namespaces(self, namespaces: Optional[NamespacesType] = None) -> NamespacesType:
+        """
+        Returns an overall namespace map for DetaElement and its descendants,
+        resolving prefix redefinitions.
+
+        :param namespaces: builds the namespace map starting over the dictionary provided.
+        """
+        namespaces = copy(namespaces) if namespaces is not None else {}
+        nsmap = None
+
+        for elem in self.iter():
+            if nsmap is elem.nsmap:
+                continue
+            else:
+                nsmap = elem.nsmap
+                for prefix, uri in nsmap.items():
+                    if not prefix:
+                        if not uri:
+                            continue
+                        elif '' not in namespaces:
+                            if self.namespace:
+                                namespaces[prefix] = uri
+                                continue
+                        elif namespaces[''] == uri:
+                            continue
+                        prefix = 'default'
+
+                    while prefix in namespaces:
+                        if namespaces[prefix] == uri:
+                            break
+                        match = re.search(r'(\d+)$', prefix)
+                        if match:
+                            index = int(match.group()) + 1
+                            prefix = prefix[:match.span()[0]] + str(index)
+                        else:
+                            prefix += '0'
+                    else:
+                        namespaces[prefix] = uri
+
+        return namespaces
+
     def validate(self, use_defaults: bool = True,
                  namespaces: Optional[NamespacesType] = None,
                  max_depth: Optional[int] = None) -> None:
@@ -227,11 +295,10 @@ class DataElement(MutableSequence['DataElement']):
             raise XMLSchemaValueError("%r has no schema bindings" % self)
 
         kwargs: Dict[str, Any] = {
+            'namespaces': self.get_namespaces(namespaces),
             'converter': DataElementConverter,
             'use_defaults': use_defaults,
         }
-        if namespaces:
-            kwargs['namespaces'] = namespaces
         if isinstance(max_depth, int) and max_depth >= 0:
             kwargs['max_depth'] = max_depth
 
@@ -255,6 +322,7 @@ class DataElement(MutableSequence['DataElement']):
         :raises: :exc:`XMLSchemaValidationError` if the object is invalid \
         and ``validation='strict'``.
         """
+        kwargs['namespaces'] = self.get_namespaces(kwargs.get('namespaces'))
         if 'converter' not in kwargs:
             kwargs['converter'] = DataElementConverter
 
@@ -270,11 +338,44 @@ class DataElement(MutableSequence['DataElement']):
 
     to_etree = encode
 
-    def tostring(self, indent: str = '', max_lines: Optional[int] = None,
-                 spaces_for_tab: int = 4) -> Any:
-        """Serializes the data element tree to an XML source string."""
-        root, errors = self.encode(validation='lax')
-        return etree_tostring(root, self.nsmap, indent, max_lines, spaces_for_tab)
+    def tostring(self, namespaces: Optional[MutableMapping[str, str]] = None,
+                 indent: str = '', max_lines: Optional[int] = None,
+                 spaces_for_tab: int = 4, xml_declaration: bool = False,
+                 encoding: str = 'unicode', method: str = 'xml') -> str:
+        """
+        Serializes the data element tree to an XML source string.
+
+        :param namespaces: is an optional mapping from namespace prefix to URI. \
+        Provided namespaces are registered before serialization. Ignored if the \
+        provided *elem* argument is a lxml Element instance.
+        :param indent: the base line indentation.
+        :param max_lines: if truncate serialization after a number of lines \
+        (default: do not truncate).
+        :param spaces_for_tab: number of spaces for replacing tab characters. For \
+        default tabs are replaced with 4 spaces, provide `None` to keep tab characters.
+        :param xml_declaration: if set to `True` inserts the XML declaration at the head.
+        :param encoding: if "unicode" (the default) the output is a string, \
+        otherwise it’s binary.
+        :param method: is either "xml" (the default), "html" or "text".
+        :return: a Unicode string.
+        """
+        root, _ = self.encode(validation='lax')
+        if not hasattr(root, 'nsmap'):
+            namespaces = self.get_namespaces(namespaces)
+
+        _string = etree_tostring(
+            elem=root,
+            namespaces=namespaces,
+            indent=indent,
+            max_lines=max_lines,
+            spaces_for_tab=spaces_for_tab,
+            xml_declaration=xml_declaration,
+            encoding=encoding,
+            method=method
+        )
+        if isinstance(_string, bytes):
+            return _string.decode('utf-8')
+        return _string
 
     def _get_xpath_context(self) -> XPathContext:
         xpath_root = build_node_tree(cast(protocols.ElementProtocol, self))
@@ -325,30 +426,6 @@ class DataElement(MutableSequence['DataElement']):
         results = parser.parse(path).select_results(context)
         yield from filter(lambda x: isinstance(x, DataElement), results)
 
-    def iter(self, tag: Optional[str] = None) -> Iterator['DataElement']:
-        """
-        Creates an iterator for the data element and its subelements. If tag
-        is not `None` or '*', only data elements whose matches tag are returned
-        from the iterator.
-        """
-        if tag == '*':
-            tag = None
-        if tag is None or tag == self.tag:
-            yield self
-        for child in self._children:
-            yield from child.iter(tag)
-
-    def iterchildren(self, tag: Optional[str] = None) -> Iterator['DataElement']:
-        """
-        Creates an iterator for the child data elements. If *tag* is not `None` or '*',
-        only data elements whose name matches tag are returned from the iterator.
-        """
-        if tag == '*':
-            tag = None
-        for child in self:
-            if tag is None or tag == child.tag:
-                yield child
-
 
 class DataBindingMeta(ABCMeta):
     """Metaclass for creating classes with bindings to XSD elements."""
@@ -391,16 +468,21 @@ class DataElementConverter(XMLSchemaConverter):
     :param namespaces: a dictionary map from namespace prefixes to URI.
     :param data_element_class: MutableSequence subclass to use for decoded data. \
     Default is `DataElement`.
+    :param map_attribute_names: define if map the names of attributes to prefixed \
+    form. Defaults to `True`. If `False` the names are kept to extended format.
     """
-    __slots__ = 'data_element_class',
+    __slots__ = 'data_element_class', 'map_attribute_names'
 
     def __init__(self, namespaces: Optional[NamespacesType] = None,
                  data_element_class: Optional[Type['DataElement']] = None,
+                 map_attribute_names: bool = True,
                  **kwargs: Any) -> None:
         if data_element_class is None:
             self.data_element_class = DataElement
         else:
             self.data_element_class = data_element_class
+
+        self.map_attribute_names = map_attribute_names
         kwargs.update(attr_prefix='', text_key='', cdata_prefix='')
         super(DataElementConverter, self).__init__(namespaces, **kwargs)
 
@@ -430,7 +512,10 @@ class DataElementConverter(XMLSchemaConverter):
     def element_decode(self, data: ElementData, xsd_element: 'XsdElement',
                        xsd_type: Optional[BaseXsdType] = None, level: int = 0) -> 'DataElement':
         data_element = self.get_data_element(data, xsd_element, xsd_type)
-        data_element.attrib.update(self.map_attributes(data.attributes))
+        if self.map_attribute_names:
+            data_element.attrib.update(self.map_attributes(data.attributes))
+        elif data.attributes:
+            data_element.attrib.update(data.attributes)
 
         if (xsd_type or xsd_element.type).model_group is not None:
             for name, value, _ in self.map_content(data.content):
