@@ -15,14 +15,14 @@ import math
 from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Pattern, \
     Tuple, Union, Counter
 from elementpath import XPath2Parser, ElementPathError, XPathToken, XPathContext, \
-    translate_pattern, datatypes
+    ElementNode, translate_pattern, datatypes
 
 from ..exceptions import XMLSchemaTypeError, XMLSchemaValueError
 from ..names import XSD_QNAME, XSD_UNIQUE, XSD_KEY, XSD_KEYREF, XSD_SELECTOR, XSD_FIELD
 from ..translation import gettext as _
 from ..helpers import get_qname, get_extended_qname
 from ..aliases import ElementType, SchemaType, NamespacesType, AtomicValueType
-from ..xpath import iter_schema_nodes
+from .exceptions import XMLSchemaNotBuiltError
 from .xsdbase import XsdComponent
 from .attributes import XsdAttribute
 
@@ -45,12 +45,9 @@ XSD_IDENTITY_XPATH_SYMBOLS = frozenset((
 ))
 
 
-# XSD identities use a restricted parser and a context for iterate element
-# references. The XMLSchemaProxy is not used for the specific selection of
-# fields and elements and the XSD fields are got at first validation run.
-class IdentityXPathContext(XPathContext):
-    _iter_nodes = staticmethod(iter_schema_nodes)
-
+# XSD identities use a restricted XPath 2.0 parser. The XMLSchemaProxy is
+# not used for the specific selection of fields and elements and the XSD
+# fields are collected at first validation run.
 
 class IdentityXPathParser(XPath2Parser):
     symbol_table = {
@@ -72,8 +69,8 @@ class XsdSelector(XsdComponent):
         lazy_quantifiers=False,
         anchors=False
     )
-    token = None   # type: XPathToken
-    parser = None  # type: IdentityXPathParser
+    token: Optional[XPathToken] = None
+    parser: Optional[IdentityXPathParser] = None
 
     def __init__(self, elem: ElementType, schema: SchemaType,
                  parent: Optional['XsdIdentity']) -> None:
@@ -165,8 +162,8 @@ class XsdIdentity(XsdComponent):
     parent: 'XsdElement'
     ref: Optional['XsdIdentity']
 
-    selector = None  # type: XsdSelector
-    fields = ()      # type: Union[Tuple[()], List[XsdFieldSelector]]
+    selector: Optional[XsdSelector] = None
+    fields: Union[Tuple[()], List[XsdFieldSelector]] = ()
 
     # XSD elements bound by selector (for speed-up and for lazy mode)
     elements: Union[Tuple[()], Dict['XsdElement', Optional[IdentityCounterType]]] = ()
@@ -210,62 +207,77 @@ class XsdIdentity(XsdComponent):
                 self.fields = ref.fields
                 self.ref = ref
 
-        context = IdentityXPathContext(self.schema, item=self.parent)  # type: ignore
+        if self.selector is None:
+            return  # Do not raise, already found by meta-schema validation.
+        elif self.selector.token is None:
+            raise XMLSchemaNotBuiltError(self, "identity selector is not built")
 
+        context = XPathContext(self.schema.xpath_node, item=self.parent.xpath_node)
         self.elements = {}
-        try:
-            for e in self.selector.token.select_results(context):
-                if not isinstance(e, XsdComponent) or isinstance(e, XsdAttribute):
-                    msg = _("selector xpath expression can only select elements")
-                    self.parse_error(msg)
-                elif e.name is not None:
-                    if TYPE_CHECKING:
-                        assert isinstance(e, XsdElement)  # for mypy checks with Python 3.7
-                    self.elements[e] = None
-        except AttributeError:
-            pass
-        else:
-            if not self.elements:
-                # Try to detect target XSD elements extracting QNames
-                # of the leaf elements from the XPath expression and
-                # use them to match global elements.
 
-                qname: Any
-                for qname in self.selector.token.iter_leaf_elements():
-                    xsd_element = self.maps.elements.get(
-                        get_extended_qname(qname, self.namespaces)
-                    )
-                    if xsd_element is not None and \
-                            not isinstance(xsd_element, tuple) and \
-                            xsd_element not in self.elements:
-                        self.elements[xsd_element] = None
+        for e in self.selector.token.select_results(context):
+            if not isinstance(e, XsdComponent) or isinstance(e, XsdAttribute):
+                msg = _("selector xpath expression can only select elements")
+                self.parse_error(msg)
+            elif e.name is not None:
+                if TYPE_CHECKING:
+                    assert isinstance(e, XsdElement)  # for mypy checks with Python 3.7
+                self.elements[e] = None
+
+        if not self.elements:
+            # Try to detect target XSD elements extracting QNames
+            # of the leaf elements from the XPath expression and
+            # use them to match global elements.
+
+            qname: Any
+            for qname in self.selector.token.iter_leaf_elements():
+                xsd_element = self.maps.elements.get(
+                    get_extended_qname(qname, self.namespaces)
+                )
+                if xsd_element is not None and \
+                        not isinstance(xsd_element, tuple) and \
+                        xsd_element not in self.elements:
+                    self.elements[xsd_element] = None
 
     @property
     def built(self) -> bool:
         return not isinstance(self.elements, tuple)
 
-    def get_fields(self, elem: Union[ElementType, 'XsdElement'],
+    def get_fields(self, element_node: ElementNode,
                    namespaces: Optional[NamespacesType] = None,
                    decoders: Optional[Tuple[XsdAttribute, ...]] = None) -> IdentityCounterType:
         """
         Get fields for a schema or instance context element.
 
-        :param elem: an Element or an XsdElement
+        :param element_node: an Element or an XsdElement
         :param namespaces: is an optional mapping from namespace prefix to URI.
         :param decoders: context schema fields decoders.
         :return: a tuple with field values. An empty field is replaced by `None`.
         """
         fields: List[IdentityFieldItemType] = []
 
-        if not isinstance(elem, XsdComponent):
-            context_class = XPathContext
-        else:
-            context_class = IdentityXPathContext
+        def append_fields() -> None:
+            if isinstance(value, list):
+                fields.append(tuple(value))
+            elif isinstance(value, bool):
+                fields.append((value, bool))
+            elif not isinstance(value, float):
+                fields.append(value)
+            elif math.isnan(value):
+                fields.append(('nan', float))
+            else:
+                fields.append((value, float))
 
         result: Any
         value: Union[AtomicValueType, None]
+
         for k, field in enumerate(self.fields):
-            result = field.token.get_results(context_class(elem))  # type: ignore
+            if field.token is None:
+                msg = f"identity field {field} is not built"
+                raise XMLSchemaNotBuiltError(self, msg)
+
+            context = XPathContext(element_node)
+            result = field.token.get_results(context)
 
             if not result:
                 if decoders is not None and decoders[k] is not None:
@@ -274,20 +286,10 @@ class XsdIdentity(XsdComponent):
                         if decoders[k].type.root_type.name == XSD_QNAME:
                             value = get_extended_qname(value, namespaces)
 
-                        if isinstance(value, list):
-                            fields.append(tuple(value))
-                        elif isinstance(value, bool):
-                            fields.append((value, bool))
-                        elif not isinstance(value, float):
-                            fields.append(value)
-                        elif math.isnan(value):
-                            fields.append(('nan', float))
-                        else:
-                            fields.append((value, float))
-
+                        append_fields()
                         continue
 
-                if not isinstance(self, XsdKey) or 'ref' in elem.attrib and \
+                if not isinstance(self, XsdKey) or 'ref' in element_node.elem.attrib and \
                         self.schema.meta_schema is None and self.schema.XSD_VERSION != '1.0':
                     fields.append(None)
                 elif field.target_namespace not in self.maps.namespaces:
@@ -311,16 +313,7 @@ class XsdIdentity(XsdComponent):
                         elif isinstance(value, datatypes.QName):
                             value = value.expanded_name
 
-                    if isinstance(value, list):
-                        fields.append(tuple(value))
-                    elif isinstance(value, bool):
-                        fields.append((value, bool))
-                    elif not isinstance(value, float):
-                        fields.append(value)
-                    elif math.isnan(value):
-                        fields.append(('nan', float))
-                    else:
-                        fields.append((value, float))
+                    append_fields()
             else:
                 msg = _("%r field selects multiple values!")
                 raise XMLSchemaValueError(msg % field)
